/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000 Sakai Hiroaki.                              */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Disp 饹Υ饤֥                                             */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "DispP.h"

static XrmOptionDescRec options[] = {};
static Atom wm_delete_window;

static void die(Widget w)
{
  kill(getpid(), SIGINT);
}

static void quit(Widget w, XEvent * event, String * params, Cardinal * num)
{
  die(w);
}

static void wm_quit(Widget w, XEvent * event, String * params, Cardinal * num)
{
  if ((event->type == ClientMessage) && 
      (event->xclient.data.l[0] != wm_delete_window)) {
    XBell(XtDisplay(w), 0);
  } else {
    die(w);
  }
}

static void iconify(Widget w, XEvent * event, String * params, Cardinal * num)
{
  XIconifyWindow(XtDisplay(w), XtWindow(w), DefaultScreen(XtDisplay(w)));
}

static void pause_and_quit(Widget w, XEvent * event, String * params,
			   Cardinal * num)
{
  sleep(10);
  die(w);
}

static XtActionsRec actions[] = {
  {"quit", quit},
  {"wm_quit", wm_quit},
  {"iconify", iconify},
  {"pause_and_quit", pause_and_quit}
};

static char * translations =
"<Message>WM_PROTOCOLS: wm_quit()\n"
"Ctrl<Key>p: pause_and_quit()\n"
"Ctrl<Key>i: iconify()\n"
"Ctrl<Key>c: quit()\n"
"None<Key>q: quit()";

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* DispChange 饹Ⱥ                                             */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* DispChange 饹                                                   */
/*---------------------------------------------------------------------------*/

static DispChange DispChangeCreate(int x, int y, int width, int height,
				   Pixmap pixmap, GC color_mask)
{
  DispChange disp_change;

  disp_change = (DispChange)malloc(sizeof(_DispChange));
  if (disp_change == NULL) ErrorExit("Cannot create disp_change.");

  disp_change->x = x;
  disp_change->y = y;
  disp_change->width  = width;
  disp_change->height = height;
  disp_change->pixmap = pixmap;
  disp_change->color_mask = color_mask;

  return (disp_change);
}

/*---------------------------------------------------------------------------*/
/* DispChange 饹κ                                                   */
/*---------------------------------------------------------------------------*/

static DispChange DispChangeDestroy(DispChange disp_change)
{
  if (disp_change == NULL) return (NULL);
  free(disp_change);
  return (NULL);
}

/*===========================================================================*/
/* ѿμ                                                    */
/*===========================================================================*/

int DispGetKeyPress(Disp disp) { return (disp->key_press); }
int DispClearKeyPress(Disp disp) { return (disp->key_press = DISP_NO_PRESS); }

/*===========================================================================*/
/* ԥåޥåפ                                                      */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ԥåޥåפ                                                      */
/*---------------------------------------------------------------------------*/

static Pixmap MakePixmapFromPixmapData(Display * display, Window w,
				       char ** string)
{
  XpmAttributes attributes;
  Pixmap pix, mask;

  if (string == NULL) return (0);

  attributes.colormap = DefaultColormap(display, DefaultScreen(display));
  attributes.closeness = 40000;
  attributes.valuemask = XpmSize|XpmReturnPixels|XpmColormap|XpmCloseness;
  XpmCreatePixmapFromData(display, w, string, &pix, &mask, &attributes);
  return (pix);
}

/*===========================================================================*/
/* Drawable Υ                                                   */
/*===========================================================================*/

static int GetX(Display * display, Drawable d)
{
  Window root;
  int x, y;
  unsigned int width, height;
  unsigned int border_width;
  unsigned int depth;

  if (!d) return (0);

  XGetGeometry(display, d, &root, &x, &y, &width, &height,
	       &border_width, &depth);
  return (x);
}

static int GetY(Display * display, Drawable d)
{
  Window root;
  int x, y;
  unsigned int width, height;
  unsigned int border_width;
  unsigned int depth;

  if (!d) return (0);

  XGetGeometry(display, d, &root, &x, &y, &width, &height,
	       &border_width, &depth);
  return (y);
}

static unsigned int GetWidth(Display * display, Drawable d)
{
  Window root;
  int x, y;
  unsigned int width, height;
  unsigned int border_width;
  unsigned int depth;

  if (!d) return (0);

  XGetGeometry(display, d, &root, &x, &y, &width, &height,
	       &border_width, &depth);
  return (width);
}

static unsigned int GetHeight(Display * display, Drawable d)
{
  Window root;
  int x, y;
  unsigned int width, height;
  unsigned int border_width;
  unsigned int depth;

  if (!d) return (0);

  XGetGeometry(display, d, &root, &x, &y, &width, &height,
	       &border_width, &depth);
  return (height);
}

static unsigned int GetDepth(Display * display, Drawable d)
{
  Window root;
  int x, y;
  unsigned int width, height;
  unsigned int border_width;
  unsigned int depth;

  if (!d) return (0);

  XGetGeometry(display, d, &root, &x, &y, &width, &height,
	       &border_width, &depth);
  return (depth);
}

/*===========================================================================*/
/* ԥåޥåפ˾Υ˳̾ơԥåޥåפ֤  */
/*===========================================================================*/

static Pixmap CreateResizedPixmap(Display * display, Drawable d, Pixmap src,
				  unsigned int dst_width,
				  unsigned int dst_height)
{
  Pixmap dst;
  XImage * src_image;
  unsigned int src_width;
  unsigned int src_height;
  int x, y, old;
  unsigned long int p0;
  unsigned long int p1;
  GC gc;

  if (!src) return (0);

  if (dst_width  < 1) dst_width  = 1;
  if (dst_height < 1) dst_height = 1;

  src_width  = GetWidth( display, src);
  src_height = GetHeight(display, src);

  gc = XCreateGC(display, DefaultRootWindow(display), 0, 0);
  XSetFunction(display, gc, GXcopy);
  XSetLineAttributes(display, gc, 1, LineSolid, CapButt, JoinMiter);

  dst = XCreatePixmap(display, d, dst_width, dst_height,
		      GetDepth(display, src));

  if ((src_width == dst_width) && (src_height == dst_height)) {
    XCopyArea(display, src, dst, gc, 0, 0, dst_width, dst_height, 0, 0);
  } else if ((src_width <= dst_width) && (src_height <= dst_height)) {
    /* ξ */
    src_image = XGetImage(display, src, 0, 0, src_width, src_height,
			  AllPlanes, XYPixmap);
    for (y = 0; y < src_height; y++) {
      for (x = 0; x < src_width; x++) {
	if (x == 0) {
	  old = x;
	  p1 = XGetPixel(src_image, x, y);
	} else
	  p1 = p0;

	if (x < src_width - 1)
	  p0 = XGetPixel(src_image, x + 1, y);

	if ((p0 != p1) || (x == src_width - 1)){
	  XSetForeground(display, gc, p1);
	  XFillRectangle(display, dst, gc,
			 old * dst_width  / src_width, /* ʲڼΤ */
			 y   * dst_height / src_height,
			 ((x + 1) * dst_width  / src_width ) -
			 (old     * dst_width  / src_width),
			 ((y + 1) * dst_height / src_height) -
			 (y       * dst_height / src_height));
	  old = x + 1;
	}
      }
    }
    XDestroyImage(src_image);
  } else if ((src_width > dst_width) && (src_height <= dst_height)) {
    /* Ĺξ */
    src_image = XGetImage(display, src, 0, 0, src_width, src_height,
			  AllPlanes, XYPixmap);
    for (x = 0; x < dst_width; x++) {
      for (y = 0; y < src_height; y++) {
	if (y == 0) {
	  old = y;
	  p1 = XGetPixel(src_image, x * src_width  / dst_width, y);
	} else
	  p1 = p0;

	if (y < src_height - 1)
	  p0 = XGetPixel(src_image, x * src_width  / dst_width, y + 1);

	if ((p0 != p1) || (y == src_height - 1)){
	  XSetForeground(display, gc, p1);
	  XDrawLine(display, dst, gc,
		    x, old * dst_height / src_height,
		    x, (y + 1) * dst_height / src_height);
	  old = y + 1;
	}
      }
    }
    XDestroyImage(src_image);
  } else if ((src_width <= dst_width) && (src_height > dst_height)) {
    /* Ĺξ */
    src_image = XGetImage(display, src, 0, 0, src_width, src_height,
			  AllPlanes, XYPixmap);
    for (y = 0; y < dst_height; y++) {
      for (x = 0; x < src_width; x++) {
	if (x == 0) {
	  old = x;
	  p1 = XGetPixel(src_image, x, y * src_height / dst_height);
	} else
	  p1 = p0;

	if (x < src_width - 1)
	  p0 = XGetPixel(src_image, x + 1, y * src_height / dst_height);

	if ((p0 != p1) || (x == src_width - 1)){
	  XSetForeground(display, gc, p1);
	  XDrawLine(display, dst, gc,
		    old * dst_width / src_width, y,
		    (x + 1) * dst_width / src_width, y);
	  old = x + 1;
	}
      }
    }
    XDestroyImage(src_image);
  } else {
    /* ̾ξ */
    src_image = XGetImage(display, src, 0, 0, src_width, src_height,
			  AllPlanes, XYPixmap);			
    for (y = 0; y < dst_height; y++) {
      for (x = 0; x < dst_width; x++) {
	if (x == 0) {
	  old = x;
	  p1 = XGetPixel(src_image,
			 x * src_width  / dst_width,
			 y * src_height / dst_height);
	} else
	  p1 = p0;

	if (x < dst_width - 1)
	  p0 = XGetPixel(src_image,
			 (x + 1) * src_width  / dst_width,
			 y       * src_height / dst_height);

	if ((p0 != p1) || (x == dst_width - 1)){
	  XSetForeground(display, gc, p1);
	  XDrawLine(display, dst, gc, old, y, x + 1, y);
	  old = x + 1;
	}
      }
    }
    XDestroyImage(src_image);
  }

  XFreeGC(display, gc);

  return (dst);
}

/*===========================================================================*/
/* ԥåޥåפ˾Υ˳̾                                */
/*===========================================================================*/

static Pixmap ResizePixmap(Display * display, Drawable d, Pixmap src,
			   unsigned int dst_width, unsigned int dst_height)
{
  Pixmap dst;
  unsigned int width;
  unsigned int height;

  if (!src) return (0);

  if (dst_width  < 1) dst_width  = 1;
  if (dst_height < 1) dst_height = 1;

  width  = GetWidth( display, src);
  height = GetHeight(display, src);

  if (width == dst_width && height == dst_height) {
    dst = src;
  } else {
    dst = CreateResizedPixmap(display, d, src, dst_width, dst_height);
    XFreePixmap(display, src);
  }

  return (dst);
}

static Pixmap ResizePixmapByRatio(Display * display, Drawable d, Pixmap src,
				  int ratio_x, int ratio_y)
{
  Pixmap dst;
  unsigned int width;
  unsigned int height;

  if (ratio_x == 100 && ratio_y == 100) {
    dst = src;
  } else {
    width  = GetWidth( display, src);
    height = GetHeight(display, src);
    dst = ResizePixmap(display, d, src,
		       width * ratio_x / 100, height * ratio_y / 100);
  }
  return (dst);
}

/*===========================================================================*/
/* ޥưŪ˺                                              */
/*===========================================================================*/

static Pixmap MakeMaskPixmap(Display * display, Drawable d,
			     Pixmap src, Pixmap dst,
			     GC figure_gc, GC mask_gc,
			     unsigned int width, unsigned int height,
			     unsigned int depth)
{
  XImage * image;
  int x, y;
  unsigned long p, p1, p2, p3, p4;

  if (!src) return (0);
  if (!dst) return (0);

  image = XGetImage(display, src, 0, 0, width, height, AllPlanes, XYPixmap);

  p1 = XGetPixel(image, 0, 0);
  p2 = XGetPixel(image, width - 1, 0);
  p3 = XGetPixel(image, 0, height - 1);
  p4 = XGetPixel(image, width - 1, height - 1);

  /* ǰ¿ */
  if ((p1 == p2) && (p1 == p3) && (p1 == p4))
    p = p1;
  else if (((p1 == p2) && (p1 == p3)) ||
	   ((p1 == p3) && (p1 == p4)) ||
	   ((p1 == p2) && (p1 == p4)))
    p = p1;
  else if ((p2 == p3) && (p2 == p4))
    p = p2;
  else if ((p1 == p2) || (p1 == p3) || (p1 == p4))
    p = p1;
  else if ((p2 == p3) || (p2 == p4))
    p = p2;
  else if (p3 == p4)
    p = p3;
  else
    p = p1;

  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      if (XGetPixel(image, x, y) == p)
	XDrawPoint(display, dst, mask_gc, x, y);
      else
	XDrawPoint(display, dst, figure_gc, x, y);
    }
  }

  XDestroyImage(image);

  return (dst);
}

static Pixmap CreateMaskPixmap(Display * display, Drawable d, Pixmap src,
			       GC figure_gc, GC mask_gc)
{
  Pixmap dst;
  unsigned int width, height, depth;

  width  = GetWidth( display, src);
  height = GetHeight(display, src);
  depth  = GetDepth( display, src);

  dst = XCreatePixmap(display, d, width, height, depth);

  MakeMaskPixmap(display, d, src, dst, figure_gc, mask_gc,
		 width, height, depth);
  return (dst);
}

/* Ϥ˹ */
static Pixmap CreateMaskBitmap(Display * display, Drawable d, Pixmap src)
{
  Pixmap dst;
  unsigned int width, height, depth;
  GC black, white;

  width  = GetWidth( display, src);
  height = GetHeight(display, src);
  depth  = GetDepth( display, src);

  dst = XCreatePixmap(display, d, width, height, 1);
  black = XCreateGC(display, dst, 0, 0);
  white = XCreateGC(display, dst, 0, 0);
  XSetForeground(display, black, BlackPixel(display, DefaultScreen(display)));
  XSetForeground(display, white, WhitePixel(display, DefaultScreen(display)));

  MakeMaskPixmap(display, d, src, dst, black, white, width, height, 1);

  XFreeGC(display, black);
  XFreeGC(display, white);

  return (dst);
}

/* Ϥ */
static Pixmap CreateInvertMaskBitmap(Display * display, Drawable d, Pixmap src)
{
  Pixmap dst;
  unsigned int width, height, depth;
  GC black, white;

  width  = GetWidth( display, src);
  height = GetHeight(display, src);
  depth  = GetDepth( display, src);

  dst = XCreatePixmap(display, d, width, height, 1);
  black = XCreateGC(display, dst, 0, 0);
  white = XCreateGC(display, dst, 0, 0);
  XSetForeground(display, black, BlackPixel(display, DefaultScreen(display)));
  XSetForeground(display, white, WhitePixel(display, DefaultScreen(display)));

  MakeMaskPixmap(display, d, src, dst, white, black, width, height, 1);

  XFreeGC(display, black);
  XFreeGC(display, white);

  return (dst);
}

/*===========================================================================*/
/* μ                                                                  */
/*===========================================================================*/

static unsigned long int DispGetColor(Disp disp, char * color_name)
{
  XColor c0, c1;
  XAllocNamedColor(disp->display, disp->colormap, color_name, &c0, &c1);
  return (c0.pixel);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Disp DispCreate(int * argcp, char * argv[], int ratio)
{
  Disp disp;

  disp = (Disp)malloc(sizeof(_Disp));
  if (disp == NULL) ErrorExit("Cannot allocate memory for display.");

  if (ratio <= 0) disp->ratio = 100;
  else disp->ratio = ratio;

  disp->game_window_width  =
    CHARACTER_SIZE * GAME_WINDOW_WIDTH  * disp->ratio / 100;
  disp->game_window_height =
    CHARACTER_SIZE * GAME_WINDOW_HEIGHT * disp->ratio / 100;

  disp->text_window_width  = FONT_WIDTH * TEXT_WINDOW_WIDTH;
  disp->text_window_height = disp->game_window_height;

  disp->game_pixmap_width  = CHARACTER_SIZE * MAP_WIDTH  * disp->ratio / 100;
  disp->game_pixmap_height = CHARACTER_SIZE * MAP_HEIGHT * disp->ratio / 100;

  disp->text_pixmap_width  = disp->text_window_width;
  disp->text_pixmap_height = disp->text_window_height;

  disp->display_width  = disp->game_window_width + disp->text_window_width;
  disp->display_height = disp->game_window_height;

  disp->toplevel = XtAppInitialize(&(disp->app_context), "Nadar",
				   NULL, 0, argcp, argv, NULL, NULL, 0);

  XtVaSetValues(disp->toplevel, XtNinput, True, NULL);
  XtVaSetValues(disp->toplevel, XtNtitle, "N.A.D.A.R.", NULL);
  XtVaSetValues(disp->toplevel, XtNwidth    , disp->display_width , NULL);
  XtVaSetValues(disp->toplevel, XtNminWidth , disp->display_width , NULL);
  XtVaSetValues(disp->toplevel, XtNmaxWidth , disp->display_width , NULL);
  XtVaSetValues(disp->toplevel, XtNheight   , disp->display_height, NULL);
  XtVaSetValues(disp->toplevel, XtNminHeight, disp->display_height, NULL);
  XtVaSetValues(disp->toplevel, XtNmaxHeight, disp->display_height, NULL);
#if 0
  disp->canvas = XtCreateWidget("Canvas", labelWidgetClass, disp->toplevel,
				NULL, 0);
#endif
  disp->key_press = DISP_NO_PRESS;
  disp->text_line = 0;

  disp->changes = SubstanceListCreate();

  return (disp);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Disp DispDestroy(Disp disp)
{
  if (disp == NULL) return (NULL);
#if 0
  if (disp->canvas) XtDestroyWidget(disp->canvas);
#endif
  if (disp->toplevel) XtDestroyWidget(disp->toplevel);

  free(disp);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* ٥ȥϥɥ                                                          */
/*---------------------------------------------------------------------------*/

static int keysym_to_keypress(KeySym key_sym)
{
  int press;

  switch (key_sym) {
  case XK_i     :   press = DISP_UP_PRESS; break;
  case XK_Up    :   press = DISP_UP_PRESS; break;
  case XK_k     :   press = DISP_DOWN_PRESS; break;
  case XK_Down  :   press = DISP_DOWN_PRESS; break;
  case XK_j     :   press = DISP_LEFT_PRESS; break;
  case XK_Left  :   press = DISP_LEFT_PRESS; break;
  case XK_l     :   press = DISP_RIGHT_PRESS; break;
  case XK_Right :   press = DISP_RIGHT_PRESS; break;
  case XK_Shift_L : press = DISP_BUTTON1_PRESS; break;
  case XK_z :       press = DISP_BUTTON1_PRESS; break;
  case XK_x :       press = DISP_BUTTON2_PRESS; break;
  case XK_Alt_L :   press = DISP_BUTTON2_PRESS; break;
  case XK_c :       press = DISP_BUTTON3_PRESS; break;
  default :         press = DISP_NO_PRESS; break;
  }
  return (press);
}

static void key_handler(Widget w, XtPointer p, XEvent * event,
			Boolean * dispatch)
{
  Disp disp = (Disp)p;
  KeySym key_sym;
  int press;
  XEvent ev;

#if 0
  while (1) {
    if ((event->type == KeyPress) || (event->type == KeyRelease)) {
      key_sym = XKeycodeToKeysym(disp->display, event->xkey.keycode, 0);
      press = keysym_to_keypress(key_sym);
      if (press != DISP_NO_PRESS) {
	if (event->type == KeyPress) disp->key_press |= press;
	else                         disp->key_press &= ~press;
      }
    }
    if (!XtAppPending(disp->app_context)) return;
    XtAppNextEvent(disp->app_context, &ev);
    event = &ev;
  }
#else
  if ((event->type == KeyPress) || (event->type == KeyRelease)) {
    key_sym = XKeycodeToKeysym(disp->display, event->xkey.keycode, 0);
    press = keysym_to_keypress(key_sym);
    if (press != DISP_NO_PRESS) {
      if (event->type == KeyPress) disp->key_press |= press;
      else                         disp->key_press &= ~press;
    }
  }
#endif

  return;
}

/*---------------------------------------------------------------------------*/
/* ݡ                                                              */
/*---------------------------------------------------------------------------*/

static void expose_handler(Widget w, XtPointer p, XEvent * event,
			   Boolean * dispatch)
{
  Disp disp = (Disp)p;

  XCopyArea(disp->display, disp->text_pixmap, XtWindow(disp->toplevel),
	    disp->copy_by_copy,
	    0, 0, disp->text_window_width, disp->text_window_height,
	    disp->game_window_width, 0);
}

/*---------------------------------------------------------------------------*/
/* ޥɥΥ줿顤                            */
/*---------------------------------------------------------------------------*/

static void leave_window_handler(Widget w, XtPointer p, XEvent * event,
				 Boolean * dispatch)
{
  Disp disp = (Disp)p;

  disp->key_press = DISP_NO_PRESS;
}

/*---------------------------------------------------------------------------*/
/* ꥢ饤                                                                */
/*---------------------------------------------------------------------------*/

static int DispRealize(Disp disp)
{
  int i;
  Display * d;
  Window w;
  int r;

  XtRealizeWidget(disp->toplevel);
  while (!XtIsRealized(disp->toplevel)) { /* None */ }

  disp->display = XtDisplay(disp->toplevel);
  disp->root_window = DefaultRootWindow(disp->display);

  d = disp->display;
  w = XtWindow(disp->toplevel);
  r = disp->ratio;

  disp->colormap = DefaultColormap(d, DefaultScreen(d));
  disp->depth = DefaultDepth(d, DefaultScreen(d));

  /*
  disp->canvas_window = XtWindow(disp->canvas);
  disp->canvas_window = XtWindow(disp->toplevel);
  */

#if 0
  disp->game_window = XCreateSimpleWindow(d, w, 0, 0,
					  disp->game_window_width,
					  disp->game_window_height,
					  0,
					  BlackPixel(d, DefaultScreen(d)),
					  BlackPixel(d, DefaultScreen(d)));

  disp->text_window = XCreateSimpleWindow(d, w,
					  disp->game_window_width, 0,
					  disp->text_window_width,
					  disp->text_window_height,
					  0,
					  BlackPixel(d, DefaultScreen(d)),
					  BlackPixel(d, DefaultScreen(d)));
#endif

  disp->game_pixmap =
    XCreatePixmap(d, w, disp->game_pixmap_width, disp->game_pixmap_height,
		  disp->depth);
  disp->text_pixmap =
    XCreatePixmap(d, w, disp->text_pixmap_width, disp->text_pixmap_height,
		  disp->depth);

  disp->copy_by_copy = XCreateGC(d, w, 0, 0);
  disp->copy_by_and  = XCreateGC(d, w, 0, 0);
  disp->copy_by_or   = XCreateGC(d, w, 0, 0);
  disp->copy_by_xor  = XCreateGC(d, w, 0, 0);
  disp->gc_black     = XCreateGC(d, w, 0, 0);
  disp->gc_white     = XCreateGC(d, w, 0, 0);

  XSetForeground(d, disp->gc_black, BlackPixel(d, DefaultScreen(d)));
  XSetBackground(d, disp->gc_black, BlackPixel(d, DefaultScreen(d)));
  XSetForeground(d, disp->gc_white, WhitePixel(d, DefaultScreen(d)));
  XSetBackground(d, disp->gc_white, WhitePixel(d, DefaultScreen(d)));

  XSetFunction(d, disp->gc_black, GXcopy);
  XSetFunction(d, disp->gc_white, GXcopy);

  XSetFunction(d, disp->copy_by_copy, GXcopy);
  XSetFunction(d, disp->copy_by_and,  GXand);
  XSetFunction(d, disp->copy_by_or,   GXor);
  XSetFunction(d, disp->copy_by_xor,  GXxor);

  /* եȤγ */
  disp->font = XLoadFont(d, "8x16");
  disp->gc_font = XCreateGC(d, w, 0, 0);
  XSetFont(d, disp->gc_font, disp->font);
  XSetFunction(d, disp->gc_font, GXcopy);
  XSetBackground(d, disp->gc_font, BlackPixel(d, DefaultScreen(d)));
  XSetForeground(d, disp->gc_font, WhitePixel(d, DefaultScreen(d)));

  /* 顼GC */
  for (i = 0; i < DISP_COLOR_NAME_NUMBER; i++) {
    disp->gc_color[i]      = XCreateGC(d, w, 0, 0);
    XSetForeground(d, disp->gc_color[i],
		   DispGetColor(disp, disp_color_name[i]));
    XSetFunction(d, disp->gc_color[i], GXcopy);
    XSetFont(d, disp->gc_color[i], disp->font);
  }

  /* ԥåޥåפɤ߹ */
  disp->icon_pixmap = MakePixmapFromPixmapData(d, w, icon_xpm);
  disp->icon_mask = CreateInvertMaskBitmap(d, w, disp->icon_pixmap);
  disp->canon_pixmap = MakePixmapFromPixmapData(d, w, canon_xpm);
  disp->missile_pixmap = MakePixmapFromPixmapData(d, w, missile_xpm);
  for (i = 0; i < TANK_ANGLE_NUMBER; i++)
    disp->tank_pixmap[i] = MakePixmapFromPixmapData(d, w, tank_pixmap_name[i]);
  for (i = 0; i < MAP_PIXMAP_NUMBER; i++)
    disp->map_pixmap[i] = MakePixmapFromPixmapData(d, w, map_pixmap_name[i]);
  for (i = 0; i < EXPLOSION_NUMBER_NUMBER; i++)
    disp->explosion_pixmap[i] =
      MakePixmapFromPixmapData(d, w, explosion_pixmap_name[i]);

  /* ԥåޥåפΥꥵ */
  disp->icon_pixmap = ResizePixmap(d, w, disp->icon_pixmap,
				   ICON_SIZE, ICON_SIZE);
  if (r != 100) {
    disp->canon_pixmap =
      ResizePixmapByRatio(d, w, disp->canon_pixmap, r, r);
    disp->missile_pixmap =
      ResizePixmapByRatio(d, w, disp->missile_pixmap, r, r);
    for (i = 0; i < TANK_ANGLE_NUMBER; i++)
      disp->tank_pixmap[i] =
	ResizePixmapByRatio(d, w, disp->tank_pixmap[i], r, r);
    for (i = 0; i < MAP_PIXMAP_NUMBER; i++)
      disp->map_pixmap[i] =
	ResizePixmapByRatio(d, w, disp->map_pixmap[i], r, r);
    for (i = 0; i < EXPLOSION_NUMBER_NUMBER; i++)
      disp->explosion_pixmap[i] =
	ResizePixmapByRatio(d, w, disp->explosion_pixmap[i], r, r);
  }

  XFillRectangle(d, w, disp->gc_black,
		 0, 0, disp->display_width, disp->display_height);

#if 0
  for (i = 0; i < DISP_COLOR_NAME_NUMBER; i++) {
    disp->tank_color_mask[i] = XCreatePixmap(d, w,
					     TANK_SIZE * r / 100,
					     TANK_SIZE * r / 100,
					     disp->depth);
    XFillRectangle(d, disp->tank_color_mask[i], disp->gc_color[i],
		   0, 0, TANK_SIZE * r / 100, TANK_SIZE * r / 100);
  }
#endif

  disp->tank_color_mask_work = XCreatePixmap(d, w,
					     TANK_SIZE * r / 100,
					     TANK_SIZE * r / 100,
					     disp->depth);

  /* ٥ȥϥɥϿ */
  XtAddEventHandler(disp->toplevel,
		    KeyPressMask | KeyReleaseMask,
		    False, key_handler, disp);
  XtAddEventHandler(disp->toplevel,
		    ExposureMask,
		    False, expose_handler, disp);
  XtAddEventHandler(disp->toplevel,
		    LeaveWindowMask | FocusChangeMask,
		    False, leave_window_handler, disp);

  /*  */
  XtVaSetValues(disp->toplevel, XtNiconPixmap, disp->icon_pixmap, NULL);
  XtVaSetValues(disp->toplevel, XtNiconMask  , disp->icon_mask  , NULL);

  /*  */
  XtAppAddActions(disp->app_context, actions, XtNumber(actions));

  /* ȥ󥹥졼 */
  XtOverrideTranslations(disp->toplevel,
			 XtParseTranslationTable(translations));

  /* ɥޥ͡㤫 f.delete ؤб */
  wm_delete_window = XInternAtom(d, "WM_DELETE_WINDOW", False);
  XSetWMProtocols(d, XtWindow(disp->toplevel), &wm_delete_window, 1);

  return (0);
}

/*---------------------------------------------------------------------------*/
/*                                                                   */
/*---------------------------------------------------------------------------*/

#if 0 /* ץѤ */
static Boolean DispMain(XtPointer p)
{
  Disp disp = (Disp)p;
  (*(disp->func))(disp->proc_data);
  return (FALSE);
}

void DispStart(Disp disp, int (*func)(), void * proc_data)
{
  disp->func = func;
  disp->proc_data = proc_data;
  DispRealize(disp);

  XtAppAddWorkProc(disp->app_context, DispMain, disp);
  XtAppMainLoop(disp->app_context);
}
#else /* ॢȤѤ */
static void DispMain(XtPointer p, XtIntervalId * interval)
{
  Disp disp = (Disp)p;
  int i, time_interval;

  i = (*(disp->func))(disp->proc_data);
  time_interval = i ? 5 : DISPLAY_INTERVAL;

  disp->interval_id = XtAppAddTimeOut(disp->app_context,
				      time_interval, DispMain, disp);
  return ;
}

void DispStart(Disp disp, int (*func)(), void * proc_data)
{
  disp->func = func;
  disp->proc_data = proc_data;
  DispRealize(disp);

  disp->interval_id = XtAppAddTimeOut(disp->app_context,
				      DISPLAY_INTERVAL, DispMain, disp);

  XtAppMainLoop(disp->app_context);
}
#endif

/*===========================================================================*/
/* Ѵؿ                                                                */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ǥץ쥤Υꥢ                                                      */
/*---------------------------------------------------------------------------*/

int DispClearDisplay(Disp disp)
{
  XEvent ev;

#if 0
  XFillRectangle(disp->display, disp->game_window, disp->gc_black, 0, 0,
		 disp->game_window_width, disp->game_window_height);
  XFillRectangle(disp->display, disp->text_window, disp->gc_black, 0, 0,
		 disp->text_window_width, disp->text_window_height);
#endif
  XFillRectangle(disp->display, disp->game_pixmap, disp->gc_black, 0, 0,
		 disp->game_pixmap_width, disp->game_pixmap_height);
  XFillRectangle(disp->display, disp->text_pixmap, disp->gc_black, 0, 0,
		 disp->text_pixmap_width, disp->text_pixmap_height);

  disp->key_press = DISP_NO_PRESS;
  disp->text_line = 0;

  while (!SubstanceListIsEmpty(disp->changes)) {
    SubstanceListSetCurrentToStart(disp->changes);
    SubstanceListDeleteSubstanceToPrev(disp->changes);
  }

  while (XtAppPending(disp->app_context)) {
    XtAppNextEvent(disp->app_context, &ev);
  }

  DispFlush(disp);

  return (0);
}

/*---------------------------------------------------------------------------*/
/*  ԥåޥåפ                                                     */
/*---------------------------------------------------------------------------*/

int DispDrawPixmap(Disp disp, int x, int y, int width, int height,
		   Pixmap pixmap, GC color_mask, int f)
{
  DispChange disp_change;

  if (color_mask) {
#if 0
    XCopyArea(disp->display, color_mask, disp->tank_color_mask_work,
	      disp->copy_by_copy,
	      0, 0, width, height, 0, 0);
#else
    XFillRectangle(disp->display, disp->tank_color_mask_work, color_mask,
		   0, 0, width, height);
#endif
    XCopyArea(disp->display, pixmap, disp->tank_color_mask_work,
	      disp->copy_by_and,
	      0, 0, width, height, 0, 0);
    XCopyArea(disp->display, disp->tank_color_mask_work, disp->game_pixmap,
	      disp->copy_by_xor,
	      0, 0, width, height, x, y);
  } else {
    XCopyArea(disp->display, pixmap, disp->game_pixmap, disp->copy_by_xor,
	      0, 0, width, height, x, y);
  }

  if (f) {
    disp_change = DispChangeCreate(x, y, width, height, pixmap, color_mask);
    SubstanceListInsertSubstanceToEnd(disp->changes, disp_change,
				      (SubstanceDestructor)DispChangeDestroy);
  }

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ޥåץ饯                                                    */
/*---------------------------------------------------------------------------*/

int DispDrawMap(Disp disp, int x, int y, int state)
{
  DispDrawPixmap(disp,
		 ConvertXFromMapToDisplay(x, disp->ratio),
		 ConvertYFromMapToDisplay(y, disp->ratio),
		 CHARACTER_SIZE * disp->ratio / 100,
		 CHARACTER_SIZE * disp->ratio / 100,
		 disp->map_pixmap[state],
		 NULL,
		 0);
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ֤                                                                */
/*---------------------------------------------------------------------------*/

int DispDrawTank(Disp disp, double x, double y, int angle, int color_number)
{
  DispDrawPixmap(disp,
		 ConvertXFromPositionToDisplay(x, disp->ratio)
		 - TANK_SIZE * disp->ratio / 200,
		 ConvertYFromPositionToDisplay(y, disp->ratio)
		 - TANK_SIZE * disp->ratio / 200,
		 TANK_SIZE * disp->ratio / 100,
		 TANK_SIZE * disp->ratio / 100,
		 disp->tank_pixmap[angle],
#if 0
		 disp->tank_color_mask[color_number % DISP_COLOR_NAME_NUMBER],
#else
		 disp->gc_color[color_number % DISP_COLOR_NAME_NUMBER],
#endif
		 1);
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ˤƤ                                                                */
/*---------------------------------------------------------------------------*/

int DispDrawCanon(Disp disp, double x, double y)
{
  DispDrawPixmap(disp,
		 ConvertXFromPositionToDisplay(x, disp->ratio)
		 - CANON_SIZE * disp->ratio / 200,
		 ConvertYFromPositionToDisplay(y, disp->ratio)
		 - CANON_SIZE * disp->ratio / 200,
		 CANON_SIZE * disp->ratio / 100,
		 CANON_SIZE * disp->ratio / 100,
		 disp->canon_pixmap,
		 NULL,
		 1);
  return (0);
}

/*---------------------------------------------------------------------------*/
/* ߥ                                                            */
/*---------------------------------------------------------------------------*/

int DispDrawMissile(Disp disp, double x, double y)
{
  DispDrawPixmap(disp,
		 ConvertXFromPositionToDisplay(x, disp->ratio)
		 - MISSILE_SIZE * disp->ratio / 200,
		 ConvertYFromPositionToDisplay(y, disp->ratio)
		 - MISSILE_SIZE * disp->ratio / 200,
		 MISSILE_SIZE * disp->ratio / 100,
		 MISSILE_SIZE * disp->ratio / 100,
		 disp->missile_pixmap,
		 NULL,
		 1);

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ȯ                                                                */
/*---------------------------------------------------------------------------*/

int DispDrawExplosion(Disp disp, double x, double y, int radius)
{
  DispDrawPixmap(disp,
		 ConvertXFromPositionToDisplay(x, disp->ratio)
		 - EXPLOSION_SIZE * disp->ratio / 200,
		 ConvertYFromPositionToDisplay(y, disp->ratio)
		 - EXPLOSION_SIZE * disp->ratio / 200,
		 EXPLOSION_SIZE * disp->ratio / 100,
		 EXPLOSION_SIZE * disp->ratio / 100,
		 disp->explosion_pixmap[radius],
		 NULL,
		 1);

  return (0);
}

/*---------------------------------------------------------------------------*/
/* եå                                                                */
/*---------------------------------------------------------------------------*/

int DispFlush(Disp disp)
{
  XFlush(disp->display);
  return (0);
}

/*---------------------------------------------------------------------------*/
/* 褷饯κ                                                  */
/*---------------------------------------------------------------------------*/

int DispCleanGamePixmap(Disp disp)
{
  DispChange change;

  while (!SubstanceListIsEmpty(disp->changes)) {
    SubstanceListSetCurrentToStart(disp->changes);
    change = (DispChange)SubstanceListGetSubstance(disp->changes);
    DispDrawPixmap(disp, change->x, change->y, change->width, change->height,
		   change->pixmap, change->color_mask, 0);
    SubstanceListDeleteSubstanceToPrev(disp->changes);
  }

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ǥץ쥤򥳥ԡ                                                  */
/*---------------------------------------------------------------------------*/

int DispDrawDisplay(Disp disp, double x, double y)
{
#ifdef PEEPING
  XCopyArea(disp->display, disp->game_pixmap, XtWindow(disp->toplevel),
	    disp->copy_by_copy,
	    ConvertXFromPositionToDisplay(x, disp->ratio)
	    - disp->game_window_width  / 2,
	    ConvertYFromPositionToDisplay(y, disp->ratio)
	    - disp->game_window_height / 2,
	    disp->game_window_width, disp->game_window_height, 0, 0);
#else

#if 0
  XCopyArea(disp->display, disp->game_pixmap, disp->game_window,
	    disp->copy_by_copy,
	    0, 0, disp->game_window_width, disp->game_window_height, 0, 0);
#else
  XCopyArea(disp->display, disp->game_pixmap, XtWindow(disp->toplevel),
	    disp->copy_by_copy,
	    0, 0, disp->game_window_width, disp->game_window_height, 0, 0);
#endif

#endif /* PEEPING */

  DispFlush(disp);

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ƥȤΥꥢ                                                          */
/*---------------------------------------------------------------------------*/

int DispClearText(Disp disp)
{
#if 0
  XFillRectangle(disp->display, disp->text_window, disp->gc_black, 0, 0,
		 disp->text_window_width, disp->text_window_height);
#endif
  XFillRectangle(disp->display, disp->text_pixmap, disp->gc_black, 0, 0,
		 disp->text_pixmap_width, disp->text_pixmap_height);

  disp->text_line = 0;

  return (0);
}

/*---------------------------------------------------------------------------*/
/* ƥȤ                                                            */
/*---------------------------------------------------------------------------*/

int DispDrawText(Disp disp, int color_number, char * string)
{
  int i;

  for (i = 0; string[i] != '\n' && string[i] != '\0'; i++) { /* None */ }

  XDrawString(disp->display, disp->text_pixmap,
	      disp->gc_color[color_number % DISP_COLOR_NAME_NUMBER],
              0, FONT_OFFSET + disp->text_line * FONT_HEIGHT, string, i);
#if 0
  XCopyArea(disp->display, disp->text_pixmap, disp->text_window,
	    disp->copy_by_copy,
	    0, 0, disp->text_window_width, disp->text_window_height, 0, 0);
#else
  XCopyArea(disp->display, disp->text_pixmap, XtWindow(disp->toplevel),
	    disp->copy_by_copy,
	    0, 0, disp->text_window_width, disp->text_window_height,
	    disp->game_window_width, 0);
#endif
  (disp->text_line)++;

  return (0);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
