/*****************************************************************************/
/* N.A.D.A.R. - A network tank battle game.                                  */
/*                                                                           */
/* N.A.D.A.R. Copyright (c) 2000 Sakai Hiroaki.                              */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* N.A.D.A.R. is a network tank battle game. You get into a tank,            */
/* operate it, and destroy other tanks to survive!                           */
/*****************************************************************************/

/*****************************************************************************/
/* Canon 饹(ˤѤΥ饹)Υ饤֥                        */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "CanonP.h"

/*****************************************************************************/
/* ؿ                                                                */
/*****************************************************************************/

/*===========================================================================*/
/* ѿμ                                                          */
/*===========================================================================*/

double CanonGetX(Canon canon) { return (canon->x); }
double CanonGetY(Canon canon) { return (canon->y); }
int CanonGetState(Canon canon)        { return (canon->state); }
int CanonSetState(Canon canon, int s) { return (canon->state = s); }
Tank CanonGetTank(Canon canon) { return (canon->tank); }

/*===========================================================================*/
/* ȯȽ                                                                */
/*===========================================================================*/

static int CanonCheckExplosion(Canon canon, Map map, SubstanceList tank_list,
			       SubstanceList explosion_list)
{
  Tank tank;
  Explosion explosion;
  double x, y, d, size;

  if (MapGetState(map,
		  ConvertXFromPositionToMap(canon->x),
		  ConvertYFromPositionToMap(canon->y))) {
    MapChangeState(map,
		   ConvertXFromPositionToMap(canon->x),
		   ConvertYFromPositionToMap(canon->y), 0);
    canon->state = CANON_EXPLOSION;
    return (canon->state);
  }

  size = (double)EXPLOSION_SIZE / CHARACTER_SIZE * 0.5 * 1.2;

  for (SubstanceListSetCurrentToStart(tank_list);
       !SubstanceListIsCurrentEndEdge(tank_list);
       SubstanceListMoveCurrentToNext(tank_list)) {
    tank = (Tank)SubstanceListGetSubstance(tank_list);
    x = canon->x - TankGetX(tank);
    y = canon->y - TankGetY(tank);
    d = x * x + y * y;
    if (d < size * size) {
      canon->state = CANON_EXPLOSION;
      return (canon->state);
    }
  }

  for (SubstanceListSetCurrentToStart(explosion_list);
       !SubstanceListIsCurrentEndEdge(explosion_list);
       SubstanceListMoveCurrentToNext(explosion_list)) {
    explosion = (Explosion)SubstanceListGetSubstance(explosion_list);
    if (ExplosionCheck(explosion, canon->x, canon->y)) {
      canon->state = CANON_EXPLOSION;
      return (canon->state);
    }
  }

  return (canon->state);
}

/*===========================================================================*/
/* ֥ȤȺ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ֥Ȥ                                                        */
/*---------------------------------------------------------------------------*/

Canon CanonCreate(double x, double y, double vx, double vy, Tank tank,
		  int state)
{
  Canon canon;

  canon = (Canon)malloc(sizeof(_Canon));
  if (canon == NULL) ErrorExit("Cannot create canon.");

  canon->x = x + vx * CANON_OFFSET;
  canon->y = y + vy * CANON_OFFSET;
  canon->vx = vx;
  canon->vy = vy;
  canon->tank = tank;
  canon->state = state;

  TankIncCanon(tank);

  return (canon);
}

/*---------------------------------------------------------------------------*/
/* ֥Ȥκ                                                        */
/*---------------------------------------------------------------------------*/

Canon CanonDestroy(Canon canon)
{
  if (canon == NULL) return (NULL);

  TankDecCanon(canon->tank);
  free(canon);

  return (NULL);
}

/*===========================================================================*/
/* ư                                                                  */
/*===========================================================================*/

static int CanonMove(Canon canon)
{
  canon->x += canon->vx;
  canon->y += canon->vy;

  return (0);
}

int CanonAct(Canon canon, Stage stage)
{
  switch (canon->state) {
  case CANON_WAIT :
    break;
  case CANON_ACTIVE :
    CanonMove(canon);
    CanonCheckExplosion(canon,
			StageGetMap(stage),
			StageGetTankList(stage),
			StageGetExplosionList(stage));
    break;
  case CANON_EXPLOSION :
    break;
  case CANON_DESTROY :
    break;
  default :
    break;
  }
  return (canon->state);
}

/*===========================================================================*/
/* ǡ                                                            */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ѥǡκ                                                        */
/*---------------------------------------------------------------------------*/

char * CanonMakeDrawData(Canon canon, char * p)
{
  if (!canon) return (NULL);
  if (!p    ) return (NULL);

  if (canon->state != CANON_ACTIVE) return (p);

  sprintf(p, CANON_DRAW "\t%3.1f %3.1f\n", canon->x, canon->y);

  while (*(p++) != '\n') { /* None */ }
  *p = '\0';

  return (p);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
