/*****************************************************************************/
/* Path.c - A library for network communication.                             */
/*                                                                           */
/* Path.c Copyright (c) 2000 Sakai Hiroaki.                                  */
/* All Rights Reserved.                                                      */
/*****************************************************************************/

/*****************************************************************************/
/* Path 饹(̿ϩΥ饹)                                               */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                   */
/*****************************************************************************/

#include "PathP.h"

/*****************************************************************************/
/* ᥽åɤ(˸ʤ)                                      */
/*****************************************************************************/

/* None */

/*****************************************************************************/
/* ᥽åɤ(˸)                                        */
/*****************************************************************************/

/*===========================================================================*/
/* ե饰λѤΥ᥽åɤ                                            */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ̿ϩ³                                                          */
/*---------------------------------------------------------------------------*/

int PathIsConnected(Path p)
{ return (p->connection == CONNECTION_CONNECTED); }
int PathIsDisconnected(Path p)
{ return (p->connection == CONNECTION_DISCONNECTED); }

/*---------------------------------------------------------------------------*/
/* ѥξ֤μ                                                      */
/*---------------------------------------------------------------------------*/

int PathIsRecvStateNodata(Path p) { return (p->recv_state == STATE_NODATA); }
int PathIsRecvStateActive(Path p) { return (p->recv_state == STATE_ACTIVE); }
int PathIsRecvStateDone(Path p)   { return (p->recv_state == STATE_DONE); }

/*---------------------------------------------------------------------------*/
/* ѥξ֤μ                                                      */
/*---------------------------------------------------------------------------*/

int PathIsSendStateNodata(Path p) { return (p->send_state == STATE_NODATA); }
int PathIsSendStateActive(Path p) { return (p->send_state == STATE_ACTIVE); }
int PathIsSendStateDone(Path p)   { return (p->send_state == STATE_DONE); }

/*===========================================================================*/
/* ֥Ȥ                                                        */
/*===========================================================================*/

Path PathNew(int fd)
{
  Path path;

  path = (Path)malloc(sizeof(_Path));

  path->sock = NULL;
  path->file_descriptor = fd;
  path->connection      = CONNECTION_CONNECTED;
  path->recv_state      = STATE_NODATA;
  path->send_state      = STATE_NODATA;
  FD_ZERO(&(path->default_mask));
  FD_SET(fd, &(path->default_mask));

  return (path);
}

Path PathCreateToAcceptRequestsFromClients(char * hostname,
					   unsigned short int port,
					   int backlog)
{
  Socket s;
  Path path;

  s = SocketCreateToAccept(hostname, port, backlog);
  if (!s) ErrorExit("Cannot create socket to accept.");

  path = PathNew(SocketGetSocket(s));
  if (!path) ErrorExit("Cannot create a new path.");

  path->sock = s;

  return (path);
}

Path PathCreateToCommunicateWithClient(Path path)
{
  return (PathNew(SocketAccept(path->sock)));
}

Path PathCreateToCommunicateWithServer(char * hostname,
				       unsigned short int port)
{
  Socket s;
  Path path;

  s = SocketCreateToConnect(hostname, port);
  if (!s) return (NULL);
  path = PathNew(SocketGetSocket(s));
  path->sock = s;

  return (path);
}

/*===========================================================================*/
/* ֥Ȥξõ                                                        */
/*===========================================================================*/

Path PathDestroy(Path path)
{
  if (path == NULL) return (NULL);

  SocketDestroy(path->sock);
  /* close(path->file_descriptor); */ /* ɤ close() 뤫 */

  free(path);

  return (NULL);
}

/*===========================================================================*/
/* ǡΥ᥽åɤ                                              */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ʸΥ᥽å                                                    */
/*---------------------------------------------------------------------------*/

char * PathSendCharacters(Path p, char * s, int len)
{
  if (p == NULL) return (NULL);
  if (s == NULL) return (NULL);

  if (len <= 0) len = strlen(s);
  if (send(p->file_descriptor, s, len, 0) == -1) {
    p->send_state = STATE_ERROR;
    return (NULL);
  }

  /* p->send_state = STATE_DONE; */

  return (s);
}

/*---------------------------------------------------------------------------*/
/* ͤΥ᥽å                                                      */
/*---------------------------------------------------------------------------*/

int PathSendIntegerValue(Path p, int value)
{
  char buf[32];
  if (p == NULL) return (-1);
  sprintf(buf, "%d", value);
  if (PathSendCharacters(p, buf, 0) == NULL) return (-1);
  return (value);
}

double PathSendDoubleValue(Path p, double value)
{
  char buf[32];
  if (p == NULL) return (0.0);
  sprintf(buf, "%e", value);
  if (PathSendCharacters(p, buf, 0) == NULL) return (0.0);
  return (value);
}

/*===========================================================================*/
/* ǡμΥ᥽åɤ                                              */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ʸμΥ᥽å                                                    */
/*---------------------------------------------------------------------------*/

char * PathRecvCharacters(Path p, char * buffer, int size)
{
  int len;
  if (p == NULL) return (NULL);
  if ((len = recv(p->file_descriptor, buffer, size - 1, 0)) == -1) {
    p->recv_state = STATE_ERROR;
    return (NULL);
  }
  buffer[len] = '\0';
  p->recv_state = STATE_DONE;
  return (buffer);
}

/*---------------------------------------------------------------------------*/
/* ͤμΥ᥽å                                                      */
/*---------------------------------------------------------------------------*/

int PathRecvIntegerValue(Path p)
{
  char buffer[32];
  if (p == NULL) return (-1);
  if (PathRecvCharacters(p, buffer, 32) == NULL) return (-1);
  return (atoi(buffer));
}

double PathRecvDoubleValue(Path p)
{
  char buffer[32];
  if (p == NULL) return (0.0);
  if (PathRecvCharacters(p, buffer, 32) == NULL) return (0.0);
  return (atof(buffer));
}

/*===========================================================================*/
/* ̿ϩξ֤Ĵ٤ʤ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* ѥξ֤Ĵ٤ʤ                                                */
/*---------------------------------------------------------------------------*/

int PathCheckRecvState(Path p)
{
  fd_set recv_mask;
  static struct timeval timeout = {0, 0};

  if (p == NULL) return (STATE_NODATA);
  recv_mask = p->default_mask;
  select(p->file_descriptor + 1, &recv_mask, NULL, NULL, &timeout);

  if (FD_ISSET(p->file_descriptor, &recv_mask))
    p->recv_state = STATE_ACTIVE;
  else
    p->recv_state = STATE_NODATA;

  return (p->recv_state);
}

/*---------------------------------------------------------------------------*/
/* ѥξ֤Ĵ٤ʤ                                                */
/*---------------------------------------------------------------------------*/

int PathCheckSendState(Path p)
{
  fd_set send_mask;
  static struct timeval timeout = {0, 0};

  if (p == NULL) return (STATE_NODATA);
  send_mask = p->default_mask;
  select(p->file_descriptor + 1, NULL, &send_mask, NULL, &timeout);

  if (FD_ISSET(p->file_descriptor, &send_mask))
    p->send_state = STATE_ACTIVE;
  else
    p->send_state = STATE_NODATA;

  return (p->send_state);
}

/*****************************************************************************/
/* ޤ                                                                  */
/*****************************************************************************/

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
